/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.io.*;
import java.nio.charset.Charset;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.*;
import java.util.*;
import java.util.logging.Level;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

abstract class AbstractSQLTool extends AbstractTool {
    private static final Pattern productNameDefinition = Pattern.compile("^(equals|contains)[ \\t]+(case|nocase)[ \\t]+([^ \\t]+)[ \\t]+([^ \\t][^\\t]*)$");
    protected final Properties connectionProperties=new Properties();
    protected final File connectionDir;
    protected final File sqlDir;
    protected final List<String[]> productNames = new ArrayList<>(10);
    protected Properties sqlProperties = null;
    private static final Pattern replaceSQL=Pattern.compile("\\[\\[([^\\[\\]]+)\\]\\]");
    private Statement currentStatement = null;

    static Cipher cipher(int mode, String secret) {
        try {
            // We simply use a digest to create a 16 byte secret. Don't care about MD5 weakness
            byte[] secretBytes = MessageDigest.getInstance("MD5").digest(secret.getBytes("UTF-8"));
            Cipher cipher=Cipher.getInstance("AES/CBC/PKCS5Padding");
            cipher.init(mode, new SecretKeySpec(secretBytes, "AES"),new IvParameterSpec(secretBytes));
            return cipher;
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        } catch (NoSuchPaddingException e) {
            throw new RuntimeException(e);
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        } catch (InvalidKeyException e) {
            throw new RuntimeException(e);
        } catch (InvalidAlgorithmParameterException e) {
            throw new RuntimeException(e);
        }

    }
    protected static  String replaceSQL(String stmt, Properties replacements) {
        StringBuffer result = new StringBuffer(stmt.length());
        Matcher matcher=replaceSQL.matcher(stmt);
        while (matcher.find())
            matcher.appendReplacement(result, replacements.getProperty(matcher.group(1),""));
        matcher.appendTail(result);
        return result.toString();
    }

    private Collection<String> stmts(Properties props, String domain, String defaultFilename, String rdbms) throws IOException   {
        List<String> result=new ArrayList<>();
        String prefix = domain == null ? "sql." : (domain + ".sql.");
        String stmt = props.getProperty(prefix + "stmt");
        if(stmt!=null) {
            result.add(replaceSQL(stmt, this.replacements));
            return result;
        }
            String sqlFileName =  props.getProperty(prefix + "file", defaultFilename);
            if (sqlFileName==null)
                return null;
            File sqlFile=new File(sqlDir,sqlFileName);
            LOG.fine("searching sqlFile "+sqlFile.getPath());
            if (sqlFile.exists()) {
                Properties scriptProperties = new Properties();
                loadScriptProperties(scriptProperties,rdbms );
                try (SQLScriptReader reader = new SQLScriptReader(new BufferedReader(new InputStreamReader(new FileInputStream(sqlFile), props.getProperty(prefix + "encoding", "UTF-8"))), scriptProperties)) {
                    for (stmt = reader.readStatement(); stmt != null; stmt = reader.readStatement()) {
                        if (!stmt.trim().isEmpty())
                            result.add(replaceSQL(stmt, this.replacements));
                    }
                }
            } else
                throw new FileNotFoundException(sqlFile.getPath());
        return result;

    }
    protected Collection<String> stmts(String domain, String defaultFilename, String rdbms) throws IOException   {
        return stmts(taskProperties, domain, defaultFilename, rdbms);
    }

    private String stmt(Properties props, String domain, String defaultFilename) throws IOException   {
        String prefix = domain == null ? "sql." : (domain + ".sql.");
        String stmt = props.getProperty(prefix + "stmt");
        if (stmt == null) {

            String sqlFileName =  props.getProperty(prefix + "file", defaultFilename);
            if (sqlFileName==null)
                return null;
            File sqlFile=new File(sqlDir,sqlFileName);
            LOG.fine("searching sqlFile "+sqlFile.getPath());
            if (sqlFile.exists()) {
                try (BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(sqlFile), props.getProperty(prefix + "encoding", "UTF-8")))) {
                    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                        if (stmt == null)
                            stmt = line;
                        else
                            stmt += "\n" + line;
                    }
                }
            } else
                throw new FileNotFoundException(sqlFile.getPath());
        }
        return replaceSQL(stmt, this.replacements);

    }
    protected String stmt(String domain, String defaultFilename) throws IOException   {
        return stmt(taskProperties, domain, defaultFilename);
    }
	protected BaseFormat[] getFormats(ResultSetMetaData md) throws SQLException {
		BaseFormat[] result=new BaseFormat[md.getColumnCount()];
		for(int col=0;col<result.length;col++) 
			result[col] = BaseFormat.valueOf(md.getColumnType(col + 1), md.getPrecision(col + 1), md.getScale(col + 1), md.getColumnTypeName(col + 1), md.getColumnName(col + 1), sqlProperties, taskProperties);
		return result;
	}
    static File getConnectionDir() {
        String name=System.getProperty("connection.dir");
        File connectionDir = name==null?null:new File(name);
        if (connectionDir != null && !connectionDir.isAbsolute()) {
            String home=System.getProperty("connection.home");
            if (home == null)
                connectionDir = connectionDir.getAbsoluteFile();
            else
                connectionDir = new File(new File(home).getAbsoluteFile(),name).getAbsoluteFile();
        }
        if (connectionDir != null && !connectionDir.isDirectory()) {
            LOG.warning("Ignoring SystemProperty connection.dir: <"+connectionDir.getPath()+"> is not an existing directory.");
            connectionDir = null;
        }
        return connectionDir;
    }
    static String getSecret(File connectionDir) throws IOException {
        String filename=System.getProperty("secret.file","secret.properties");
        File f = new File(filename);
        if(!f.isAbsolute())
            f = new File(connectionDir, filename);
        if (!f.isFile())
            throw new FileNotFoundException(f.getPath());
        Properties props = new Properties();
        props.load(new FileInputStream(f));
        String secret = props.getProperty("secret");
        if (secret == null)
            throw new NullPointerException("Property secret not found in File "+f.getAbsolutePath());
        return secret;
    }
    protected AbstractSQLTool(String taskname, String connection, String directory, boolean supportSourceDir)throws IOException {
    	super(taskname, directory, supportSourceDir);
        this.connectionDir = getConnectionDir();
        File connectionDir = this.connectionDir;
        File f = null;
        if (connection != null) {
            if (connectionDir == null && !new File(connection).isAbsolute()) {
                String home=System.getProperty("connection.home");
                if (home != null)
                    connectionDir = new File(home).getAbsoluteFile();
                if (connectionDir != null && !connectionDir.isDirectory()) {
                	LOG.warning("Ignoring SystemProperty connection.home: <"+connectionDir.getPath()+"> is not an existing directory.");
                    connectionDir = null;
                }
            }
            LOG.fine("connectionDir="+connectionDir.getPath());
            if (!connectionDir.isDirectory())
                LOG.warning("connectionDir is not a directory");
            f=new File(connectionDir, connection);
            if (!(f.isFile() || connection.toLowerCase().endsWith(".properties"))) {
                LOG.fine(f.getPath()+" is not a file. try with suffix .properties");
                f=new File(connectionDir, connection+".properties");
            }
            if (!(f.isFile() || connection.toLowerCase().endsWith(".props"))) {
                LOG.fine(f.getPath()+" is not a file. try with suffix .props");
                f=new File(connectionDir, connection+".props");
            }
            if (f.isFile()) {
                LOG.fine(f.getPath()+" found");
                connectionProperties.load(new FileInputStream(f));
            }
            else
                throw new FileNotFoundException(f.getPath());
        } else {
            String name=System.getProperty("connection.properties");
            f=new File(connectionDir, name);
            if (f.isFile())
                connectionProperties.load(new FileInputStream(f));
            else
                throw new FileNotFoundException(f.getPath());
        }
        // Now check for encrypted password
        String encryptedPassword = connectionProperties.getProperty("encrypted-password");
        if (encryptedPassword != null) {
            connectionProperties.remove("encrypted-password");
            String password = connectionProperties.getProperty("password");
            if (password == null) {
                try {
                    password = new String(cipher(Cipher.DECRYPT_MODE,getSecret(connectionDir)).doFinal(Base64.getDecoder().decode(encryptedPassword)));
                    connectionProperties.setProperty("password", password);
                } catch (IllegalBlockSizeException e) {
                    throw new RuntimeException(e);
                } catch (BadPaddingException e) {
                    throw new RuntimeException(e);
                }
            }
        }

        String name=System.getProperty("sql.dir");
        File sqlDir= name==null?null:new File(name).getAbsoluteFile();
        if (sqlDir != null && !sqlDir.isDirectory()) {
        	LOG.warning("Ignoring SystemProperty sql.dir: <"+name+"> is not an existing directory.");
            sqlDir = configDir;
        } else if (sqlDir == null)
            sqlDir = configDir;
        this.sqlDir=sqlDir;
        loadProductNames(configDir);
    }
    protected AbstractSQLTool(String taskname, String connection, String directory) throws IOException {
    	this(taskname,connection,directory,false);
    }
    /*
    protected AbstractSQLTool(String taskname, Properties connectionProperties, File configDir, Properties replacements)throws IOException {
    }
*/
    private void loadProductNames(File configDir) throws IOException {
        try (Reader reader=AbstractTool.fileOrResourceUTF8(configDir, null, "Productnames.txt", "com/enitsys/sqltools/Productnames.txt", "Productnames.txt", "com/enitsys/sqltools/Productnames.txt")) {
            if (reader==null)
                throw new NullPointerException("reader");
            try (BufferedReader definitions = new BufferedReader(reader)) {
                Matcher matcher;
                for(String definition=definitions.readLine();definition!=null;definition=definitions.readLine()) {
                    matcher = productNameDefinition.matcher(definition);
                    if (matcher.matches())
                        productNames.add(new String[]{matcher.group(1),matcher.group(2),matcher.group(3),matcher.group(4).trim()});
                }
            }
        }
    }
    private String toRDBMS(String databaseProductname) {
        if (databaseProductname == null)
            return null;
        for (String[] definitions : productNames) {
            if (definitions[0].equals("equals") && definitions[1].equals("case")) {
                if (definitions[3].equals(databaseProductname))
                    return definitions[2];
            } else if (definitions[0].equals("equals")) {
                if (definitions[3].equalsIgnoreCase(databaseProductname))
                    return definitions[2];
            }
            else if (definitions[0].equals("contains") && definitions[1].equals("case")) {
                if (databaseProductname.contains(definitions[3]))
                    return definitions[2];
            } else if (definitions[0].equals("contains")) {
                if (databaseProductname.toUpperCase(Locale.ROOT).contains(definitions[3].toUpperCase(Locale.ROOT)))
                    return definitions[2];
            }
        }
        return databaseProductname;
    }
    protected final Connection createConnection(boolean log) throws SQLException,IOException {
        String driver = connectionProperties.getProperty("driver");
        if(driver != null) {
            try {
                Class.forName(driver);
            } catch(ClassNotFoundException e) {
                if (log)
            	    LOG.warning("defined JDBC driver class not found: "+driver);
            }
        }
        if (log)
            LOG.info("connecting to <"+connectionProperties.getProperty("url")+">...");
        if (connectionProperties.getProperty("url").contains(":oracle:")&&!connectionProperties.containsKey("v$session.program"))
        	connectionProperties.setProperty("v$session.program", getClass().getSimpleName());
        
        Connection connection = DriverManager.getConnection(connectionProperties.getProperty("url"), connectionProperties);
        if (connection!=null) {
            String val=connectionProperties.getProperty("schema");
            if(val!=null)
            	connection.setSchema(val);
            val=connectionProperties.getProperty("catalog");
            if(val!=null)
            	connection.setCatalog(val);
            if (sqlProperties==null) {
                DatabaseMetaData meta = connection.getMetaData();
                String rdbms = toRDBMS(meta.getDatabaseProductName());
                LOG.fine("db.productname="+ rdbms);
                taskProperties.setProperty("db.productname", rdbms);
                replacements.setProperty("db.productname", rdbms);
                String dbversion = meta.getDatabaseProductVersion();
                LOG.fine("db.productversion="+ dbversion);
                replacements.setProperty("db.productversion", dbversion);
                sqlProperties = new Properties();
                loadSQLProperties(sqlProperties, rdbms);
            }
        }
        return connection;
    }
    protected final Connection createConnection() throws SQLException,IOException {
        return createConnection(true);
    }

    public abstract void run() throws SQLException, IOException;
    protected synchronized final void setCurrentStatement(Statement currentStatement) {
        this.currentStatement=currentStatement;
    }
    public void kill() {
        Statement currentStatement;
        synchronized (this) {
            currentStatement=this.currentStatement;
        }
        if (currentStatement == null)
            return;
        try {
            currentStatement.cancel();
        } catch (SQLException e) {
            LOG.log(Level.WARNING,"Failed to cancel current SQL Statement.", e);
        } finally {
            setCurrentStatement(null);
        }
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        kill();
    }
}
