/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.nio.charset.StandardCharsets;
import java.sql.*;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

abstract class AbstractTool {
	static final Logger LOG = Logger.getLogger(AbstractTool.class.getPackage().getName());
    protected final Properties taskProperties=new Properties();
    protected final Properties replacements;
    protected final File configDir;
    protected final File destDir;
    protected final File sourceDir;

    protected final String taskname;

    private static final Pattern replace=Pattern.compile("\\[([^\\[\\]]+)\\]");
    protected static  String replace(String stmt, Properties replacements) {
        if (stmt == null)
            return null;
        StringBuffer result = new StringBuffer(stmt.length());
        Matcher matcher=replace.matcher(stmt);
        while (matcher.find())
            matcher.appendReplacement(result, replacements.getProperty(matcher.group(1)));
        matcher.appendTail(result);
        return result.toString();
    }

    protected static OutputStream loggerOutputStream(Level level) {
        return new LoggerOutputStream(LOG,level);
    }

    static <T, X1 extends T, X2 extends T> T nvl(X1 x1, X2 x2) {
    	return x1!=null?x1:x2;
    }
    private void backwardsCompatible(Properties props, String oldKey, String newKey) {
    	if (!props.containsKey(newKey) && props.containsKey(oldKey))
    		props.setProperty(newKey, props.getProperty(oldKey));
    }
    private Properties loadTaskProperties(File configDir, String fileName) throws IOException {
        Properties props=new Properties();
        try (Reader reader=fileUTF8(configDir, fileName)) {
            props.load(reader);
        }
        String imp=props.getProperty("config$import");
        if(imp!=null) {
            for (Map.Entry<Object, Object> e : loadTaskProperties(configDir, imp).entrySet())
                if (!props.containsKey(e.getKey()))
                    props.put(e.getKey(), e.getValue());
        }
        props.remove("config$import");
        String imps=props.getProperty("config$imports");
        if(imps!=null) {
            for (StringTokenizer st = new StringTokenizer(imps, ",;"); st.hasMoreTokens(); ) {
                imps= st.nextToken();
                if(imps!=null) {
                    imp=props.getProperty("config$import."+imps, imps);
                    imp=imp.trim();
                    if (!imp.isEmpty()) {
                        for (Map.Entry<Object, Object> e : loadTaskProperties(configDir, imp).entrySet())
                            if (!props.containsKey(e.getKey()))
                                props.put(e.getKey(), e.getValue());
                    }
                    props.remove("config$import."+imps);
                }
            }
        }
        props.remove("config$imports");
        return props;
    }
    protected AbstractTool(String taskname, String directory, boolean supportSourceDir)throws IOException {
        backwardsCompatible(System.getProperties(), "config.dir", "configuration.dir");
        String name=System.getProperty("configuration.dir");
        File configDir= name==null?null:new File(name).getAbsoluteFile();
        if (configDir != null && !configDir.isDirectory()) {
        	LOG.warning("Ignoring SystemProperty configuration.dir: <"+name+"> is not an existing directory.");
            configDir = null;
        }
        this.configDir=configDir;

        this.taskname=taskname;
        if (new File(configDir, taskname+".properties").isFile())
            this.taskProperties.putAll(loadTaskProperties(configDir, taskname+".properties"));
        else
            this.taskProperties.putAll(System.getProperties());
        backwardsCompatible(this.taskProperties, "header", "csv.header");
        backwardsCompatible(this.taskProperties, "separator", "csv.separator");
        backwardsCompatible(this.taskProperties, "filename", "out.file");
        backwardsCompatible(this.taskProperties, "encoding", "out.encoding");
        backwardsCompatible(this.taskProperties, "fetchsize", "sql.fetchsize");
        backwardsCompatible(this.taskProperties, "temp-directory", "out.tempdir");

        this.replacements=new Properties(taskProperties);
        this.replacements.putAll(System.getProperties());
        
        File destDir = directory==null?null:new File(directory).getAbsoluteFile();
        /*
        if (destDir != null)
            System.out.println("Found out.dir as Argument: "+destDir.getPath());
         */
        if (destDir != null && !destDir.isDirectory()) {
            if (!destDir.exists()) {
                if (!destDir.mkdirs())
                    throw new IOException("Failed to create directory <"+ destDir.getPath()+">.");
            }
            else
                throw new IOException("<"+ destDir.getPath()+"> is not a directory.");
        } else if (destDir == null) {
            name=System.getProperty("out.dir");
            destDir = name==null?null:new File(name).getAbsoluteFile();
            /*
            if (destDir != null)
                System.out.println("Found out.dir as System Property: "+destDir.getPath());
             */
            if (destDir != null && !destDir.isDirectory()) {
                if (!destDir.exists()) {
                    if (!destDir.mkdirs())
                        throw new IOException("Failed to create directory <"+ destDir.getPath()+">.");
                }
                else
                    throw new IOException("<"+ destDir.getPath()+"> is not a directory.");
            } else if (destDir == null){
                name = taskProperties.getProperty("out.dir");
                destDir = name == null ? null : new File(name).getAbsoluteFile();
                /*
                if (destDir != null)
                    System.out.println("Found out.dir as Task Property: "+destDir.getPath());
                 */
                if (destDir != null && !destDir.isDirectory()) {
                    if (!destDir.exists()) {
                        if (!destDir.mkdirs())
                            throw new IOException("Failed to create directory <" + destDir.getPath() + ">.");
                    } else
                        throw new IOException("<" + destDir.getPath() + "> is not a directory.");
                }
            }
        }
        this.destDir = destDir;

        File sourceDir = null;
        if (supportSourceDir) {
        	name=System.getProperty("in.dir");
        	sourceDir = name==null?null:new File(name).getAbsoluteFile();
        	if (sourceDir != null && sourceDir.exists()) {
        		if (!sourceDir.isDirectory())
                    throw new IOException("<"+ sourceDir.getPath()+"> is not a directory.");
            } else if (sourceDir == null){
                name = taskProperties.getProperty("in.dir");
                sourceDir = name == null ? null : new File(name).getAbsoluteFile();
            	if (sourceDir != null && sourceDir.exists()) {
            		if (!sourceDir.isDirectory())
                        throw new IOException("<"+ sourceDir.getPath()+"> is not a directory.");
                } else if (sourceDir !=null)
                	sourceDir = null;
            } else
            	sourceDir=null;
        }
        this.sourceDir = sourceDir;


    }
    public File inputFile(String inputFilename) throws IOException {
		String inFilename=inputFilename==null?taskProperties.getProperty("in.file"):inputFilename;
		if(inFilename==null)
			throw new NullPointerException("inputFilename");
		File inFile = new File(inFilename);
		if (!inFile.isAbsolute())
			inFile=new File(sourceDir, inFilename).getAbsoluteFile();
		if(inFile.exists()) {
			if(!inFile.isFile())
                throw new IOException("<" + inFile.getPath() + "> is not an accessible file.");
			return inFile;
		}
		else
            throw new IOException("<" + inFile.getPath() + "> does not exist.");

    }
    static Reader fileUTF8(File dir, String fileName) throws IOException {
        File file = new File(fileName);
        if(!file.isAbsolute() && dir!= null && dir.isDirectory())
            file=new File(dir, fileName);
        file=file.getAbsoluteFile();
        if(!file.isFile())
            return null;
        return new InputStreamReader(new FileInputStream(file), StandardCharsets.UTF_8);
    }
    static Reader fileOrResourceUTF8(File dir, String fileName, String defaultFileName, String defaultResource, String fallbackFileName, String fallbackResource) throws IOException {
    	fileName=fileName==null?defaultFileName:fileName;
		
		File file = new File(fileName);
		if(!file.isAbsolute() && dir!= null && dir.isDirectory())
			file=new File(dir, fileName);
		file=file.getAbsoluteFile();
		
		InputStream is = null;
		if (file.isFile())
			is =new FileInputStream(file);
		else {
			is = AbstractTool.class.getClassLoader().getResourceAsStream(fileName);
			if (is==null)
				is=AbstractTool.class.getClassLoader().getResourceAsStream(defaultResource);
		}
		if (is==null) {
			file = new File(fallbackFileName);
			if(dir!= null && dir.isDirectory())
				file=new File(dir, fileName);
			file=file.getAbsoluteFile();
			if (file.isFile())
				is =new FileInputStream(file);
			else {
				is = AbstractTool.class.getClassLoader().getResourceAsStream(fallbackFileName);
				if (is==null)
					is=AbstractTool.class.getClassLoader().getResourceAsStream(fallbackResource);
			}
		}
		return is==null?null:new InputStreamReader(is, StandardCharsets.UTF_8);
    }
    Reader fileOrResourceUTF8(String fileName, String defaultFileName, String defaultResource, String fallbackFileName, String fallbackResource) throws IOException {
    	 return fileOrResourceUTF8(configDir, fileName,defaultFileName,defaultResource, fallbackFileName, fallbackResource);
    }


    private void loadRDBMSProperties(Properties properties, String propertiesType, String rdbms) throws IOException {
        try (Reader reader= fileOrResourceUTF8(null, propertiesType+"."+rdbms+".properties", "com/enitsys/sqltools/"+propertiesType+"."+rdbms+".properties", propertiesType+".ansi.properties", "com/enitsys/sqltools/"+propertiesType+".ansi.properties")) {
            if (reader==null)
                throw new NullPointerException("reader");
            properties.load(reader);
        }
    }
    protected void loadDMLProperties(Properties dmlProperties, String rdbms) throws IOException {
        loadRDBMSProperties(dmlProperties,"dml",rdbms);
	}
    protected void loadDDLProperties(Properties ddlProperties, String rdbms) throws IOException {
        loadRDBMSProperties(ddlProperties,"ddl",rdbms);
	}
    protected void loadSQLProperties(Properties sqlProperties, String rdbms) throws IOException {
        loadRDBMSProperties(sqlProperties,"sql",rdbms);
    }
    protected void loadScriptProperties(Properties sqlProperties, String rdbms) throws IOException {
        loadRDBMSProperties(sqlProperties,"script",rdbms);
    }

    public abstract void run() throws SQLException, IOException;


    protected static <Tool extends AbstractTool> Tool newInstance(Class<Tool> clazz, String[] args) {
        try {
            Constructor<Tool> constructor = clazz.getDeclaredConstructor(String.class, String.class, String.class);
            return constructor.newInstance(args[0], args.length>1 ? args[1] : null, args.length>2 ? args[2] : null);
        } catch (InvocationTargetException e) {
            e.getCause().printStackTrace();
            System.err.println("configuration error");
            System.exit(2);
        } catch (Exception e) {
            e.printStackTrace();
            System.err.println("configuration error");
            System.exit(2);
        }
        return null;
    }
    protected static <Tool extends AbstractTool> void main(Class<Tool> clazz, String[] args) {
    	long start=System.currentTimeMillis();
        if(args.length==0) {
            System.err.println("missing argument <taskname>");
            System.err.println("usage: java [options] "+clazz.getName()+" <taskname> [...]");
            System.exit(1);
        }
        if("simple".equalsIgnoreCase(System.getProperty("logging.format","simple")))
        	System.setProperty("java.util.logging.SimpleFormatter.format","%5$s%6$s%n");

        String loglevel=System.getProperty("logging.level");
        if(loglevel!=null) {
        	Level level=Level.parse(loglevel);
        	Logger logger=LOG;
        	while (logger!=null) {
        		logger.setLevel(level);
        		for(Handler handler:logger.getHandlers())
        			handler.setLevel(level);
        		logger=logger.getParent();
        	}
        }
        
        if (System.getProperty("connection.home") == null) {
            if (System.getProperty("os.name", "blabla").toUpperCase().startsWith("WIN"))
                System.setProperty("connection.home", System.getenv("HOMEDRIVE") + System.getenv("HOMEPATH"));
            else
                System.setProperty("connection.home", System.getProperty("user.home"));
        }
        Package pkg = AbstractTool.class.getPackage();
        System.setProperty("sqltools.classname", clazz.getSimpleName());
        if (pkg.getSpecificationVersion()!=null)
            System.setProperty("sqltools.version", pkg.getSpecificationVersion());
        System.setProperty("sqltools.today", LocalDate.now().format(DateTimeFormatter.BASIC_ISO_DATE));
        System.setProperty("sqltools.now", LocalDateTime.now().format(DateTimeFormatter.ofPattern("yyyyMMddHHmmss")));
        LOG.info(clazz.getSimpleName() + (pkg.getSpecificationVersion()==null?"":(" ("+pkg.getSpecificationTitle()+" Version "+pkg.getSpecificationVersion()+") - "+pkg.getSpecificationVendor())));
        LOG.info(" --- Documentation: http://www.enitsys.com/SQLTools.pdf ---");

        Tool tool = AbstractTool.newInstance(clazz, args);
        try {
            tool.run();
            LOG.info(clazz.getName() +" "+args[0]+(args.length>1?(" "+args[1]):"")+(args.length>2?(" "+args[2]):"")+" succesfully completed in "+((System.currentTimeMillis()-start)/1000L)+" seconds.");
            //System.exit(0);
        } catch (IOException e) {
            e.printStackTrace();
            System.err.println("IO error");
            System.exit(3);
        } catch (SQLException e) {
            e.printStackTrace();
            System.err.println("SQL error");
            System.exit(4);
        } catch (Exception e) {
            e.printStackTrace();
            System.err.println("Unexpected error");
            System.exit(5);
        } catch (Throwable e) {
            e.printStackTrace();
            System.err.println("VM error");
            System.exit(6);
        }


    }
}
