/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.enitsys.sqltools.CSVFormat.CSVDefinition;

public final class CSV2sql extends AbstractSQLTool {
	private static final Pattern regexPrecision=Pattern.compile("\\{precision\\}");
	private static final Pattern regexScale=Pattern.compile("\\{scale\\}");
	private static final Pattern regexTablename=Pattern.compile("\\{tablename\\}");
	private static final Pattern regexColumnname=Pattern.compile("\\{columnname\\}");

	private final CSVFormat csvFormat = new CSVFormat();
	private final CSVDefinition csvDefinition;
	private final boolean header;
	private final boolean columnsByName;
	private final boolean ignoreMissingColumns;
	private final File inputFile;
	private final String tablename;
	private final Properties dmlProperties=new Properties();
	private final int batchsize;
	CSV2sql(String taskname, String connection, String inputFilename)throws IOException {
		super( taskname, connection, null, true);
		csvFormat.loadFormatDefinitions(configDir, taskProperties);
		this.inputFile=inputFile(inputFilename);
		this.tablename=System.getProperty("out.table",taskProperties.getProperty("out.table",taskname));
		this.batchsize=Integer.parseInt(taskProperties.getProperty("out.batchsize","1000"));
		this.csvDefinition=CSVDefinition.valueOf(csvFormat, taskProperties);
		String header=taskProperties.getProperty("csv.header");
		String columns=taskProperties.getProperty("columns");
		if (header==null && columns!=null)
			header="name".equalsIgnoreCase(columns)?"true":"false";
		if(header==null)
			header="false";
		if(columns==null)
			columns="true".equalsIgnoreCase(header)?"name":"position";
		this.header="true".equals(header);
		this.columnsByName="name".equalsIgnoreCase(columns);
		String mandatory=taskProperties.getProperty("columns.mandatory","false");
		this.ignoreMissingColumns=!"true".equalsIgnoreCase(mandatory);
	}
	private int maxColumnPosition() {
		int result=0;
		Pattern keyPattern=Pattern.compile("^column\\.([0-9]+)\\..+$");
		Matcher matcher;
		for(Object key : taskProperties.keySet()) {
			matcher=keyPattern.matcher((String) key);
			if(matcher.matches())
				result=Math.max(result, new Integer(matcher.group(1)));
		}
		if(columnsByName)
			return result;
		keyPattern=Pattern.compile("^map\\.[^\\.]+$");
		Pattern valuePattern=Pattern.compile("^([0-9]+)$");
		for(Map.Entry<?,?> entry : taskProperties.entrySet())
			if(keyPattern.matcher((String)entry.getKey()).matches()) {
				matcher=valuePattern.matcher((String) entry.getValue());
				if(matcher.matches())
					result=Math.max(result, new Integer(matcher.group(1)));
			}
		return result;
	}
	private List<String> getCSVColumnNames(int maxColumnPosition,BufferedReader csv) throws IOException {
		if (header) {
			String line=csv.readLine();
			if(line==null)
				return null;
			if(!columnsByName)
				return Collections.emptyList();
			List<String> result = csvDefinition.getValues(line);
			String colName;
			for(int i=0; i<result.size();i++) {
				colName=result.get(i);
				if (colName!=null&&colName.length()>0)
					result.set(i, taskProperties.getProperty("column."+colName+".name", colName));
			}
			for(int i=0; i<result.size();i++)				
				result.set(i, taskProperties.getProperty("column."+(i+1)+".name", result.get(i)));
			for(int i=result.size(); i<maxColumnPosition;i++)				
				result.add(i, taskProperties.getProperty("column."+(i+1)+".name", "COLUMN"+(i+1)));
				
			return result;
		}
		if (!columnsByName)
			return Collections.emptyList();
		List<String> result = new ArrayList<String>(maxColumnPosition);
		for(int i=0; i<maxColumnPosition;i++)				
			result.add(i, taskProperties.getProperty("column."+(i+1)+".name", "COLUMN"+(i+1)));
		return result;
	}
	
	int[] getColumnMap(ResultSetMetaData resultSetMetaData, int tableColCount, List<String> csvColumnNames, Case casesensitivity) throws SQLException {
		int[] result=new int[tableColCount];
//		boolean uppercase="true".equalsIgnoreCase(taskProperties.getProperty("columns.map-uppercase",casesensitivity.map_uppercase_default()));
		boolean automapping="true".equalsIgnoreCase(taskProperties.getProperty("columns.map-auto"));
		if(columnsByName) {
			String value;
			String key;
			Integer position;
			Map<String, Integer> map=new HashMap<String, Integer>();
			for(int i=0;i<csvColumnNames.size();i++) {
				value=csvColumnNames.get(i);
				if(value!=null)
					map.put(casesensitivity.value(value), i+1);
			}
			if (automapping) {
				for(int i=0; i<tableColCount;i++) {
					position=map.get(resultSetMetaData.getColumnName(i+1));
					if (position!=null)
						result[i] = position;
				}
			}
			for(int i=0; i<tableColCount;i++) {
				key="map."+resultSetMetaData.getColumnName(i+1);
				if(taskProperties.containsKey(key)) {
					value=taskProperties.getProperty(key);
					if(value!=null && value.trim().length()>0) {
						value=casesensitivity.value(value);
						position=map.get(value);
						if(position==null && !ignoreMissingColumns)
							throw new RuntimeException("Column "+value+" not found in Inputfile");
						if (position!=null)
							result[i] = position;
					} else
						result[i] = 0;
				}
			}
			
		} else {
			String value;
			for(int i=0; i<tableColCount;i++) {
				value=taskProperties.getProperty("map."+resultSetMetaData.getColumnName(i+1));
				if(value!=null && value.trim().length()>0)
					result[i] = Integer.parseInt(value);
			}
		}
		return result;
		
	}
	CSVFormat.ColumnFormat[] getCSVFormats(int[] columnMap, List<String> csvColumnNames) {
		List<CSVFormat.ColumnFormat> result=new ArrayList<>();
		for(int i=0;i<columnMap.length;i++) {
			if(columnMap[i]>0) {
				while(result.size() < columnMap[i])
					result.add(null);
				if(result.get(columnMap[i]-1)==null) {
					if(columnsByName)
						result.set(columnMap[i]-1, csvFormat.getColumnFormat(taskProperties.getProperty("column."+csvColumnNames.get(columnMap[i]-1)+".format","STRING")));
					else
						result.set(columnMap[i]-1, csvFormat.getColumnFormat(taskProperties.getProperty("column."+columnMap[i]+".format","STRING")));
				}
			}
		}
		return result.toArray(new CSVFormat.ColumnFormat[result.size()]);
	}
	String getInsertStmt(Case casesensitivity, String tablename, int[] columnMap, ResultSetMetaData resultSetMetaData, BaseFormat[] tableFormats, CSVFormat.ColumnFormat[] csvFormats) throws SQLException {

		StringBuilder sb=new StringBuilder(1000);
		sb.append("insert into ").append(tablename);
		boolean first=true;
		Matcher dmlColumnname=regexColumnname.matcher(dmlProperties.getProperty(casesensitivity.dml_columnname(),"{columnname}"));
		for(int i=0;i<columnMap.length;i++) {
			if(columnMap[i]>0) {
				sb.append(first?'(':',');
				sb.append(dmlColumnname.replaceAll(casesensitivity.value(resultSetMetaData.getColumnName(i+1))));
				first=false;
			}
		}
		if(first)
			return null;
		sb.append(")values");
		first=true;
		String pattern;
		for(int i=0;i<columnMap.length;i++) {
			if(columnMap[i]>0) {
				sb.append(first?'(':',');
//				LOG.fine("Map TableColumn "+(i+1) +" "+tableFormats[i]+" to CSVColumn "+columnMap[i]+" "+csvFormats[columnMap[i]-1].baseFormat);
				pattern=dmlProperties.getProperty("dml."+csvFormats[columnMap[i]-1].baseFormat.name()+"."+tableFormats[i].name(),"?");
				pattern=regexPrecision.matcher(pattern).replaceAll(String.valueOf(resultSetMetaData.getPrecision(i+1)));
				pattern=regexScale.matcher(pattern).replaceAll(String.valueOf(resultSetMetaData.getScale(i+1)));
				sb.append(pattern);
				first=false;
			}
		}
		sb.append(')');
		return sb.toString();
	}
	private int[] columnMapEntries(int[] columnMap) {
		List<Integer> ll=new ArrayList<>(columnMap.length);
		for(int i=0;i<columnMap.length;i++)
			if(columnMap[i]>0)
				ll.add(i);
		int[] result=new int[ll.size()];
		for(int i=0;i<ll.size();i++)
			result[i]=ll.get(i);
		return result;

	}
	private void process(PreparedStatement stmt, List<String[]> records, int[] counts) throws SQLException {
		try {
			for(String[] record: records) {
				for(int i=1;i<record.length;i++)
					stmt.setString(i, record[i]);
				stmt.addBatch();
			}
			stmt.executeBatch();
			counts[0]+=records.size();
			stmt.getConnection().commit();
			return;
		} catch (SQLException e) {
			//ignore
		}
		stmt.clearBatch();
		if (records.size()>100) {
			int split=records.size()/2;
			process(stmt,  records.subList(0, split), counts);
			process(stmt,  records.subList(split, records.size()), counts);
			return;
		}
		for(String[] record: records) {
			try {
				for(int i=1;i<record.length;i++)
					stmt.setString(i, record[i]);
				stmt.execute();
				counts[0]++;
				stmt.getConnection().commit();
			} catch (SQLException e) {
				LOG.warning("Failed to insert line "+record[0]+": "+e.getMessage());
				counts[1]++;
			}
		}
	}
	public void run() throws SQLException, IOException{
		
		int[] counts={0,0,0,0};
		try (LineNumberReader csv = new LineNumberReader(new InputStreamReader(new FileInputStream(inputFile),"UTF-8"))) {
			int maxPosition=maxColumnPosition();
			List<String> csvColumnNames=getCSVColumnNames(maxPosition, csv);
			if (csvColumnNames == null)
				throw new RuntimeException("File does not contain line with column names");
			try (Connection conn =createConnection()) {
				conn.setAutoCommit(false);
				loadDMLProperties(this.dmlProperties, taskProperties.getProperty("db.productname"));
				Case casesensitivity=Case.valueOf(System.getProperty("out.case",taskProperties.getProperty("out.case",dmlProperties.getProperty("dml.case","sensitive"))).trim().toLowerCase());
				String dmlTablename=regexTablename.matcher(dmlProperties.getProperty(casesensitivity.dml_tablename(),"{tablename}")).replaceAll(casesensitivity.value(tablename));
				BaseFormat[] tableFormats;
				CSVFormat.ColumnFormat[] csvFormats;
				int[] columnMap;
				String sql;
				try (Statement stmt = conn.createStatement()) {
//					System.setProperty("timestamp.execute", new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()));
					try (ResultSet rs = stmt.executeQuery("select * from "+dmlTablename+" where 1=2")) {
						ResultSetMetaData resultSetMetaData = rs.getMetaData();
						int tableColCount = resultSetMetaData.getColumnCount();
						tableFormats=getFormats(resultSetMetaData);
						columnMap=getColumnMap(resultSetMetaData, tableColCount, csvColumnNames, casesensitivity);
						csvFormats=getCSVFormats(columnMap, csvColumnNames);
						sql=getInsertStmt(casesensitivity, dmlTablename, columnMap, resultSetMetaData, tableFormats, csvFormats);
					}
					String preparation=dmlProperties.getProperty("dml.preparation");
					if(preparation!=null && !preparation.trim().isEmpty())
						stmt.execute(preparation);
				}
				int[] columnMapEntries=columnMapEntries(columnMap);
				LOG.fine("preparing sql:" +sql);
				List<String> values;
				int srcColumn;
				String[] record;
				List<String[]> records=new ArrayList<>();
				try (PreparedStatement stmt = conn.prepareStatement(sql)) {
					for(String line=csv.readLine();line!=null;line=csv.readLine()) {
						try {
							values=csvDefinition.getValues(line);
							if(values!= null) {
								record=new String[columnMapEntries.length+1];
								record[0]=String.valueOf(csv.getLineNumber());
								for(int i=0;i<columnMapEntries.length;i++) {
									srcColumn=columnMap[columnMapEntries[i]]-1;
									record[i+1]=csvFormats[srcColumn].toSQLValue(values.size()<=srcColumn?null:values.get(srcColumn));
								}
								records.add(record);
							} else {
								counts[3]++;
							}
						} catch(Exception e) {
							counts[2]++;
								// Ignore
						}
						if(records.size()>=batchsize) {
							process(stmt, records,counts);
							records=new ArrayList<>();
						}
					}
					if(records.size()>0) {
						process(stmt, records,counts);
					}
				}
			}
		}
		LOG.info("completed: "+counts[0]+" records uploaded. ");
		if(counts[1]>0)
			LOG.warning("           "+counts[1]+" records import failed");
		if(counts[2]>0)
			LOG.warning("           "+counts[2]+" records parsing failed");
		if(counts[3]>0)
			LOG.warning("           "+counts[3]+" empty lines skipped");
		
		
	}
	public static void main(String[] args) {
		/*
		System.setProperty("logging.level", "FINE");
		System.setProperty("logging.simple", "true");
		System.setProperty("configuration.dir", "test");
		System.setProperty("connection.dir", "jdbcconnection");
		
		String[] xargs=(args==null || args.length == 0)?new String[]{"Test1","sqltoolstest","src/test/Test1.txt"}:args;
		main(CSV2sql.class, xargs);
		*/
		main(CSV2sql.class, args);
	}
}
