/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

final class CSVFormat {
	private static final Set<Character> regexReserved=new HashSet<>(Arrays.asList('.','[','{','}','(',')','\\','*','+','?','|','^','$'));
	private static String toRegex(String s) {
		StringBuilder sb=new StringBuilder(2*s.length());
		for(char c : s.toCharArray()) {
			if (regexReserved.contains(c))
				sb.append('\\');
			sb.append(c);
		}
		return sb.toString();
	}
	
	private static enum Quoting {
		NEVER,
		SMART,
		NONEMPTYSTRING,
		STRING,
		NONEMPTY,
		ALLWAYS;
	}

	private static class Named {
		protected final String name;
		protected Named(String name) {
			this.name=name;
		}
	}
	static abstract class ColumnFormat extends Named {
		protected final BaseFormat baseFormat;
		protected final Pattern regex;
		@SuppressWarnings("rawtypes")
		protected Comparable minValue=null;
		@SuppressWarnings("rawtypes")
		protected Comparable maxValue=null;
		protected String minValueStr=null;
		protected String maxValueStr=null;
		ColumnFormat(String name, BaseFormat formatType, String regex) {
			super(name);
			this.baseFormat=formatType;
			this.regex=regex==null?null:Pattern.compile(regex);
//			AbstractTool.LOG.fine(this.formatType+" "+this.name);
		}
		public boolean matches(String value) {
			if (value==null || value.length()==0)
				return true;
			if (regex==null)
				return true;
			return regex.matcher(value).matches();
		}
		@SuppressWarnings({ "rawtypes", "unchecked" })
		protected void matched(Comparable value, String valueStr ) {
	   		if(minValue==null||value.compareTo(minValue) < 0) {
	   			minValue=value;
	   			minValueStr=valueStr;
	   		}
	   		if(maxValue==null||value.compareTo(maxValue) > 0) {
	   			maxValue=value;
	   			maxValueStr=valueStr;
	   		}
		}
		public abstract String toSQLValue(String value);
		@SuppressWarnings("unchecked")
		public boolean isConstant() {
			return (minValue!=null && minValue.compareTo(maxValue)==0);
		}
		@SuppressWarnings("unchecked")
		public String constantValue() {
			return (minValue!=null && minValue.compareTo(maxValue)==0)?minValueStr:null;
		}
		public String toString() {
			return baseFormat +" " +name;
		}
		public void write(PrintStream out, String prefix) {
			out.println(prefix+"format="+name);
			if(isConstant())
				out.println((prefix.startsWith("#")?"":"## ")+prefix+"value="+constantValue());
			
		}
		public abstract ColumnFormat clone();
		public String ddlFormat(Properties ddlProperties) {
			return ddlProperties.getProperty("ddl.format."+baseFormat,"VARCHAR(100)"); 
		}

	}
	
	static final class DateFormat extends ColumnFormat {
		private final DateTimeFormatter simpleDateFormat;
		private final String dtfPattern;
		DateFormat(String name, BaseFormat baseFormat, String regex, Locale locale, String sdfPattern) {
			super(name, baseFormat, regex);
			this.dtfPattern=sdfPattern;
			this.simpleDateFormat=new DateTimeFormatterBuilder().appendPattern(sdfPattern).toFormatter(locale);
		}
		public ColumnFormat clone() {
			return new DateFormat(name, baseFormat, regex.pattern(), simpleDateFormat.getLocale(),dtfPattern);
		}

		public boolean matches(String value) {
			if (!super.matches(value))
				return false;
			@SuppressWarnings("rawtypes")
			Comparable result;
	   		try {
				result=baseFormat.parse(value, simpleDateFormat);
				if (result==null)
					return false;
			} catch (DateTimeParseException e) {
				return false;
			}
	   		matched(result, value);
	   		return true;
		}
		public String toSQLValue(String value) {
			if (value==null || value.length()==0)
				return null;
	   		try {
				return baseFormat.getSQLString(baseFormat.parse(value, simpleDateFormat));
			} catch (DateTimeParseException e) {
				return null;
			}
		}
	}
	static final class NumberFormat extends ColumnFormat {
		private static final Pattern white=Pattern.compile("\\s");
		private static final Pattern dot=Pattern.compile("\\.");
		private static final Pattern comma=Pattern.compile(",");
		private final char[] normalizerFlags;
		private static String remove(Pattern p, String x) {
			return p.matcher(x).replaceAll("");
		}
		private static String comma2dot(String x) {
			return x.replace(',', '.');
		}

		private static String insert0Prefix(String x) {
			char c0=x.charAt(0);
			if(c0=='.')
				return new StringBuilder(x.length()+1).append('0').append(x).toString();
			if((c0=='+' || c0=='-') && x.charAt(1)=='.')
				return new StringBuilder(x.length()+1).append(x).insert(1, c0).toString();
			return x;
		}
		private static String normalize(char flag, String x) {
			switch (flag) {
			case 'c': return remove(comma, x);
			case 'd': return remove(dot, x);
			case 'w': return remove(white, x);
			case 'C': return comma2dot(x);
			case '0': return insert0Prefix(x);
			}
			return x;
		}
		private int minPrecision=-1;
		private int maxPrecision=0;
		private int minIntegerPlaces=-1;
		private int maxIntegerPlaces=0;
		private int minScale=-1;
		private int maxScale=0;
		NumberFormat( String name, BaseFormat formatType,String regex, String normalizerFlags) {
			super(name, formatType, regex);
			
			this.normalizerFlags=normalizerFlags==null?new char[] {}:normalizerFlags.toCharArray();
		}
		public  ColumnFormat clone() {
			return new NumberFormat(name, baseFormat, regex.pattern(), String.valueOf(normalizerFlags));
		}
		private String normalize(String x) {
			String y=x;
			for(char flag:normalizerFlags)
				y=normalize(flag, y);
			return y;
		}
		public boolean matches(String value) {
			if (value==null || value.length()==0)
				return true;
			if (regex==null)
				return true;
			if(!regex.matcher(value).matches())
				return false;
			String normalizedValue =normalize(value);
			BigDecimal result=new BigDecimal(normalizedValue);
			if(BaseFormat.INTEGER.equals(baseFormat))
				result=new BigDecimal(result.unscaledValue());
			matched(result,normalizedValue);
			int scale=result.scale();
			if(scale<0)
				throw new RuntimeException("Scale<0 not supported: value="+normalizedValue+" - scale="+scale);
			if(minScale<0 || scale<minScale)
				minScale=scale;
			if(scale>maxScale)
				maxScale=scale;
			int precision = result.precision();
			if(minPrecision<0 || precision<minPrecision)
				minPrecision=precision;
			if(precision>maxPrecision)
				maxPrecision=precision;
			int round=precision-scale;
			if(minIntegerPlaces<0 || round<minIntegerPlaces)
				minIntegerPlaces=round;
			if(round>maxIntegerPlaces)
				maxIntegerPlaces=round;
			
			return true;
		}
		public String toSQLValue(String value) {
			if (value==null || value.length()==0)
				return null;
			value=normalize(value);
			if(value.length()==0)
				return null;
			if(BaseFormat.INTEGER.equals(baseFormat))
				return baseFormat.getSQLString(new BigDecimal(normalize(value)).unscaledValue());
			return baseFormat.getSQLString(new BigDecimal(normalize(value)));
		}
		
		public String ddlFormat(Properties ddlProperties) {
			String ddl=ddlProperties.getProperty("ddl.format."+baseFormat,"VARCHAR(100)"); 
			ddl=Pattern.compile("\\{precision\\}").matcher(ddl).replaceAll(String.valueOf(maxPrecision));
			ddl=Pattern.compile("\\{scale\\}").matcher(ddl).replaceAll(String.valueOf(maxScale));
			ddl=Pattern.compile("\\{integerpart\\}").matcher(ddl).replaceAll(String.valueOf(maxIntegerPlaces));
			return ddl;
		}
		
	}
	static final class StringFormat extends ColumnFormat {
		private int minLen=0;
		private int maxLen=0;
		private int count = 0;
		
		StringFormat() {
			super("STRING", BaseFormat.STRING, null);
		}
		public boolean matches(String value) {
			if (value==null)
				throw new NullPointerException("value");
			int len=value.length();
			if(len==0)
				throw new NullPointerException("value");
			count++;
			if (minLen==0 || minLen>len)
				minLen=len;
			if (maxLen==0 || maxLen<len)
				maxLen=len;
			matched(value,value);
	   		return true;
		}
		public String toString() {
			if (isConstant())
				return "STRING length>="+maxLen+ " CONSTANT "+constantValue();
			if (count>0)
				return "STRING length>="+maxLen;
			return "ANY (ALLWAYS EMPTY)"; 
		}
		@Override
		public ColumnFormat clone() {
			return new StringFormat(); 
		}
		public String ddlFormat(Properties ddlProperties) {
			String ddl=ddlProperties.getProperty("ddl.format.STRING","VARCHAR({size})"); 
			ddl=Pattern.compile("\\{size\\}").matcher(ddl).replaceAll(String.valueOf(maxLen));
			return ddl;
		}
		public String toSQLValue(String value) {
			if (value==null || value.length()==0)
				return null;
			return value;
		}

		
	}
	static class SeparatorDefinition extends Named {
		private final String separator;
		
		SeparatorDefinition(String name, String separator) {
			super(name);
			this.separator = separator;
		}
		public String toString() {
			return "separator="+separator;
		}

	}
	static class EscapeDefinition extends Named {
		private final String escape;
		
		EscapeDefinition(String name, String escape) {
			super(name);
			this.escape = escape;
			if (escape.length()>1)
				throw new RuntimeException("EscapeDefinition "+name+" is longer than 1 Character: <"+escape+">");
		}
		public String toString() {
			return "escape="+escape;
		}

	}
	static class QuotesDefinition extends Named {
		private final String left;
		private final String right;
		QuotesDefinition(String name, String left, String right) {
			super(name);
			this.left = left;
			this.right = right;
		}
		public String toString() {
			return "left="+left+" right="+right;
		}
		

	}
	static final class Unescape {
		private final Pattern escapePattern;
		private final Pattern unescapePattern;
		private final String unescapeReplacement;
		private final String escapeReplacement;
		
		Unescape(String escape, String replacement) {
			this.escapePattern=Pattern.compile(toRegex(replacement));
			this.unescapePattern=Pattern.compile(toRegex(escape)+toRegex(replacement));
			this.escapeReplacement=Matcher.quoteReplacement(escape+replacement);
			this.unescapeReplacement=Matcher.quoteReplacement(replacement);
		}
		String unescape(String value) {
			return unescapePattern.matcher(value).replaceAll(unescapeReplacement);
		}
		String escape(String value) {
			return escapePattern.matcher(value).replaceAll(escapeReplacement);
		}
		
	}
	static class CSVDefinition implements Comparable<CSVDefinition>{
		private final SeparatorDefinition separatorDefinition;
		private final EscapeDefinition escapeDefinition;
		private final QuotesDefinition quotesDefinition;
		
		private final Quoting quoting;
		private final Pattern separator;
		private final Pattern escape;
		private final Pattern left;
		private final Pattern right;
		private final String emptyQuotes;
		private final boolean header;
		private final boolean quoteHeader;
		private final boolean rectangle;
		private final boolean suppressEmptyLines;

		private final Unescape[] replacements;
		
		private int numLines;
		private int numFirstLineColumns;
		private int numLinesNoSplit;
		private int numLinesFails;
		private int numLinesLessColumnsThanFirst;
		private int numLinesMoreColumnsThanFirst;
		private int numEscapes=0;
		private int numQuotes=0;
		private final boolean rightStartsWithEscape;
		
		private List<List<String>> lines=new ArrayList<>();
		
		
		
		public int getNumFirstLineColumns() {
			return numFirstLineColumns;
		}
		@Override
		public int compareTo(CSVDefinition o) {
			int result = o.numLines - numLines;
			if (result!=0)
				return result;
			if(numLines==0)
				return 0;
			result = numLinesNoSplit - o.numLinesNoSplit;
			if (result!=0)
				return result;
			//check rectangle
			result = ((numFirstLineColumns>1)&&(numLinesLessColumnsThanFirst+numLinesMoreColumnsThanFirst==0)?-1:0)+((o.numFirstLineColumns>1)&&(o.numLinesLessColumnsThanFirst+o.numLinesMoreColumnsThanFirst==0)?1:0);
			if (result!=0)
				return result;
			result = (escapeDefinition!=null?-1:0)+(o.escapeDefinition!=null?1:0);
			if (result!=0)
				return result;
			result = (quotesDefinition!=null?-1:0)+(o.quotesDefinition!=null?1:0);
			if (result!=0)
				return result;
			
			return 0;
		}
		private CSVDefinition(SeparatorDefinition separatorDefinition, EscapeDefinition escapeDefinition, QuotesDefinition quotesDefinition, Quoting quoting, boolean header, boolean quteHeader, boolean rectangle, boolean suppressEmptyLines) {
			this.separatorDefinition = separatorDefinition;
			this.escapeDefinition = escapeDefinition;
			this.quotesDefinition = quotesDefinition;
			this.quoting=quoting;
			this.quoteHeader = quteHeader;
			this.rightStartsWithEscape = escapeDefinition!=null && quotesDefinition!=null && escapeDefinition.escape.equals(quotesDefinition.right.substring(0, 1));
			this.header=header;
			this.rectangle = rectangle;
			this.suppressEmptyLines = suppressEmptyLines;
			this.emptyQuotes=quotesDefinition==null?null:quotesDefinition.left+quotesDefinition.right;
			this.separator=Pattern.compile(toRegex(separatorDefinition.separator));
			String regex=(escapeDefinition==null||rightStartsWithEscape)?null:toRegex(escapeDefinition.escape);
			this.escape=null;//regex==null?null:Pattern.compile("(^|[^"+regex+"])"+regex+"("+regex+regex+")*[^"+regex+"]");
			this.left=quotesDefinition==null?null:Pattern.compile(toRegex(quotesDefinition.left));
			regex=null;
			if (quotesDefinition!=null) {
				if(escapeDefinition==null) {
					regex="("+toRegex(quotesDefinition.right)+")";
				} else {
					String regesc=toRegex(escapeDefinition.escape);
					regex="(?:^|[^"+regesc+"])(?:"+regesc+regesc+")*("+toRegex(quotesDefinition.right)+")(?:$|"+toRegex(separatorDefinition.separator)+")";
				}
			}
			this.right=quotesDefinition==null?null:Pattern.compile(regex);
			
			if(escapeDefinition==null)
				this.replacements=new Unescape[] {};
			else if (rightStartsWithEscape) 
				this.replacements=new Unescape[]  {new Unescape(escapeDefinition.escape,escapeDefinition.escape)};
			else  
				this.replacements=new Unescape[]  {new Unescape(escapeDefinition.escape,escapeDefinition.escape),new Unescape(escapeDefinition.escape,quotesDefinition.right.substring(0,1))};
		}
		public String toString() {
			int columns=0;
			for(int i=0;i<lines.size();i++)
				columns+=lines.get(i).size();
			int fieldsize=0;
			for(List<String> x:lines)
				for(String s:x)
					if (s!=null)
						fieldsize+=s.length();
			
			return separatorDefinition.toString()+" "+(escapeDefinition==null?"NO ESCAPE":escapeDefinition.toString())+" "+(quotesDefinition==null?"NO QUOTES":quotesDefinition.toString())
					+ "  "+lines.size()+" lines, "+columns+" columns, "+fieldsize+ " fieldsize";
		}
		static CSVDefinition valueOf(SeparatorDefinition separatorDefinition, QuotesDefinition boundaryDefinition, EscapeDefinition escapeDefinition) {
			if(separatorDefinition==null)
				throw new NullPointerException("separatorDefinition");
			if (boundaryDefinition==null && escapeDefinition!=null)
				return null;
			if (boundaryDefinition!=null) {
				if(separatorDefinition.separator.equals(boundaryDefinition.left))
					return null;
				if(escapeDefinition!=null && boundaryDefinition.right.length()>1 
					&& escapeDefinition.escape.equals(boundaryDefinition.right.substring(0, 1)) 
					&& escapeDefinition.escape.equals(boundaryDefinition.right.substring(1, 2)))
					return null;
			}
			return new CSVDefinition(separatorDefinition, escapeDefinition, boundaryDefinition, Quoting.SMART, false, false, true, true);
		}
		static CSVDefinition valueOf(CSVFormat format, Properties properties) {
			SeparatorDefinition separator=new SeparatorDefinition(null,";");
			String prop=properties.getProperty("csv.separator");
			if(prop==null) {
				prop=properties.getProperty("csv.separator.name");
				if(prop!=null) {
					format.ensureMaps();
					separator=format.separatorDefinitionByName.get(prop);
					if(separator==null)
						throw new RuntimeException("specified separator name "+prop+" not found.");
				}
			} else
				separator=new SeparatorDefinition(null,prop);

			QuotesDefinition quotes=null;
			prop=properties.getProperty("csv.quotes.left");
			if(prop==null) {
				prop=properties.getProperty("csv.quotes.name");
				if(prop!=null && !"NONE".equalsIgnoreCase(prop)) {
					format.ensureMaps();
					quotes=format.quoteDefinitionByName.get(prop);
					if(quotes==null)
						throw new RuntimeException("specified quote name "+prop+" not found.");
				}
			} else {
				String right=properties.getProperty("csv.quotes.right");
				if(right==null)
					throw new RuntimeException("missing property csv.quotes.right.");
				
				quotes=new QuotesDefinition(null,prop,right);
			}

			EscapeDefinition escape=null;
			prop=properties.getProperty("csv.escape");
			if(prop==null) {
				prop=properties.getProperty("csv.escape.name");
				if(prop!=null && !"NONE".equalsIgnoreCase(prop)) {
					format.ensureMaps();
					escape=format.escapeDefinitionByName.get(prop);
					if(escape==null)
						throw new RuntimeException("specified escape name "+prop+" not found.");
				} else if (prop==null && quotes!=null) {
					escape=new EscapeDefinition(null,quotes.right.substring(0,1));
				}
			} else
				escape=new EscapeDefinition(null,prop.substring(0,1));

			Quoting quoting=Quoting.NEVER;
			if (quotes!=null) {
				prop=properties.getProperty("csv.quoting","SMART");
				quoting=Quoting.valueOf(prop.toUpperCase());
			}
				
			boolean header="TRUE".equalsIgnoreCase(properties.getProperty("csv.header","true"));
			boolean quoteHeader="TRUE".equalsIgnoreCase(properties.getProperty("csv.quote.header",(quoting.equals(Quoting.ALLWAYS)||quoting.equals(Quoting.NONEMPTY))?"true":"false"));
			boolean rectangle="TRUE".equalsIgnoreCase(properties.getProperty("csv.rectangle","true"));
			boolean suppresEmptyLine="TRUE".equalsIgnoreCase(properties.getProperty("csv.suppressEmptyLines","true"));

			if(escape!=null && escape.escape.equals(separator.separator))
				throw new RuntimeException("Invalid Format Specification: separator and escape must not be equal ("+separator.separator+")");
			if(escape!=null && escape.escape.equals(separator.separator.substring(0,1)))
				throw new RuntimeException("Invalid Format Specification: separator must not start with escape ("+separator.separator+")");
			if(quotes!=null && separator.separator.equals(quotes.left))
				throw new RuntimeException("Invalid Format Specification: separator and left quoting must not be equal ("+separator.separator+")");
			if(escape!=null && quotes!=null && quotes.right.length()>1 
					&& escape.escape.equals(quotes.right.substring(0, 1)) && escape.escape.equals(quotes.right.substring(1, 2)))
				throw new RuntimeException("Invalid Format Specification: right quote must not be start with doubled escape character ("+quotes.right+")");
			return new CSVDefinition(separator, escape, quotes, quoting, header, quoteHeader, rectangle,suppresEmptyLine);
		}
		public String separator() {
			return separatorDefinition.separator;
		}
		public boolean header() {
			return header;
		}
		public boolean quoteHeader() {
			return quoteHeader;
		}
		public boolean rectangle() {
			return rectangle;
		}
		public boolean suppressEmptyLines() {
			return suppressEmptyLines;
		}
		private String unescape(String value) {
			String result=value;
			for(Unescape unescape:replacements)
				result=unescape.unescape(result);
			return result;
		}
		public String quote(String value) {
			for (Unescape unescape: replacements)
				value=unescape.escape(value);
			return new StringBuilder(value.length()+quotesDefinition.left.length()+quotesDefinition.right.length())
					.append(quotesDefinition.left).append(value).append(quotesDefinition.right).toString();
		}
		String encode(String value, BaseFormat format) {
			if (value!=null && !value.isEmpty())
				value=value.replace("\n", "").replace("\r", "");
				
			if (value==null || value.isEmpty()) {
				switch(quoting) {
				case ALLWAYS:
					return emptyQuotes;
				case STRING:
					return BaseFormat.STRING.equals(format)?emptyQuotes:null;
				default:
					return null;
				}
			}
			switch(quoting) {
			case NEVER:
			return value.replace(separatorDefinition.separator, "") ;
			case ALLWAYS:
			case NONEMPTY:
				return quote(value);
			case STRING:
			case NONEMPTYSTRING:
				return BaseFormat.STRING.equals(format)?quote(value):value;
			default: //SMART
				if (value.contains(separatorDefinition.separator) || value.startsWith(quotesDefinition.left))
					return quote(value);
				return value;
			}
		}
		/*
		public List<String> getValues(String line) {
			List<String> result=new ArrayList<>();
			Matcher matcher=separator.matcher(line);
			Matcher escapeMatcher=escape==null?null:escape.matcher(line);
			
			int valStart=0;
			int valEnd=0;
			int escaped=Integer.MAX_VALUE;
			if(escapeMatcher!=null && escapeMatcher.find()) {
				escaped=escapeMatcher.end()-1;
				escapeMatcher.region(escaped+1, escapeMatcher.regionEnd());
			} 
			boolean tryUnquoted;
			boolean sepFound=false;
			int endStart=0;
			String rawValue;
			String unescapedValue;
			
			int nextStart=0;
			
			for (@SuppressWarnings("unused") int i=1;;i++) {
				tryUnquoted=true;
				if (left != null) {
					matcher.usePattern(left);
					if(matcher.lookingAt()) {
						tryUnquoted=false;
						valStart=matcher.end();
						matcher.region(matcher.end(), matcher.regionEnd());
						matcher.usePattern(right);
						for(boolean endFound=false;!endFound;) {
							endFound=matcher.find();
							if(!endFound)
								throw new RuntimeException("Position "+ matcher.regionStart() +": Closing ColumnBoundaries "+right.pattern()+" not found.");
							endStart=matcher.start();
							nextStart=matcher.end();
							while (escaped<endStart) {
								if(escapeMatcher.find()) {
									escaped=escapeMatcher.end()-1;
									escapeMatcher.region(escaped+1, escapeMatcher.regionEnd());
								} else 	
									escaped=Integer.MAX_VALUE;
							}
 if (escaped==endStart && !rightStartsWithEscape) {
								endFound=false;
								matcher.region(escaped+1, matcher.regionEnd());
							}

						}
						valEnd=endStart;
						matcher.region(nextStart, matcher.regionEnd());
					
						matcher.usePattern(separator);
						sepFound=matcher.lookingAt();
						if(!sepFound && matcher.regionStart() < matcher.regionEnd() )
							throw new RuntimeException("Position "+ matcher.regionStart() +": EOL or ColumnSeparator "+separator.pattern()+" expected.");
						numQuotes++;
					}
				} 
				if (tryUnquoted) {
					valStart=matcher.regionStart();
					matcher.usePattern(separator);
					for(boolean endFound=false;!endFound;) {
						sepFound=matcher.find();
						if(!sepFound)
							break;
						endStart=matcher.start();
						while (escaped<endStart) {
							if(escapeMatcher.find()) {
								escaped=escapeMatcher.end()-1;
								escapeMatcher.region(escaped+1, escapeMatcher.regionEnd());
							} else 
								escaped=Integer.MAX_VALUE;
						}
						endFound = (escaped != endStart);
						if(!endFound)
							matcher.region(escaped+1, matcher.regionEnd());
					}
					valEnd=sepFound?matcher.start():matcher.regionEnd();
				}
				rawValue=line.substring(valStart, valEnd);
				unescapedValue=unescape(rawValue);
				if(rawValue.length()!=unescapedValue.length())
					numEscapes++;
				result.add(unescapedValue);
				if(!sepFound)
					break;
				matcher.region(matcher.end(), matcher.regionEnd());
			}
			return result;
		}
*/
public List<String> getValues(String line) {
	List<String> result=new ArrayList<>();
	Matcher matcher=separator.matcher(line);
	
	int valStart=0;
	int valEnd=0;
	boolean tryUnquoted;
	boolean sepFound=false;
	int endStart=0;
	String rawValue;
	String unescapedValue;
	
	int nextStart=0;
	
	for (@SuppressWarnings("unused") int i=1;;i++) {
		tryUnquoted=true;
		if (left != null) {
			matcher.usePattern(left);
			if(matcher.lookingAt()) {
				tryUnquoted=false;
				valStart=matcher.end();
				matcher.region(matcher.end(), matcher.regionEnd());
				matcher.usePattern(right);
				if(!matcher.find())
					throw new RuntimeException("Position "+ matcher.regionStart() +": Closing ColumnBoundaries "+right.pattern()+" not found.");
				endStart=matcher.start(1);
				nextStart=matcher.end(1);
				valEnd=endStart;
				matcher.region(nextStart, matcher.regionEnd());
			
				matcher.usePattern(separator);
				sepFound=matcher.lookingAt();
				if(!sepFound && matcher.regionStart() < matcher.regionEnd() )
					throw new RuntimeException("Position "+ matcher.regionStart() +": EOL or ColumnSeparator "+separator.pattern()+" expected.");
				numQuotes++;
			}
		} 
		if (tryUnquoted) {
			valStart=matcher.regionStart();
			matcher.usePattern(separator);
			sepFound=matcher.find();
			if(sepFound)
				endStart=matcher.start();
			valEnd=sepFound?matcher.start():matcher.regionEnd();
		}
		rawValue=line.substring(valStart, valEnd);
		unescapedValue=unescape(rawValue);
		if(rawValue.length()!=unescapedValue.length())
			numEscapes++;
		result.add(unescapedValue);
		if(!sepFound)
			break;
		matcher.region(matcher.end(), matcher.regionEnd());
	}
	return result;
}

	void analyse(int lineNumber, String line) {
			try {
				List<String> values=getValues(line);
				if(numLines==0) {
					numFirstLineColumns=values.size();
					if(values.size()==1)
						numLinesNoSplit++;
				} else {
					if(values.size()==1)
						numLinesNoSplit++;
					else if(values.size()<numFirstLineColumns)
						numLinesLessColumnsThanFirst++;
					else if(values.size()>numFirstLineColumns)
						numLinesMoreColumnsThanFirst++;
					
				}
				numLines++;
				if (lines.size()>100)
					lines.remove(1);
				lines.add(values);
			} catch (Exception e) {
				numLinesFails++;
				numLines++;
				AbstractTool.LOG.fine("line "+ lineNumber+": "+this+" fails by "+line+" - "+e.getMessage());
			}
		}
		boolean isRemovable(boolean finalRemovement) {
			if(numLines<100 && !finalRemovement)
				return false;
			if(numLines==numLinesFails)
				return true;
			if(numLines==numLinesNoSplit)
				return true;
			if(finalRemovement) {
				if (quotesDefinition!=null && numQuotes==0)
					return true;
				if (escapeDefinition!=null && numEscapes==0)
					return true;
			}
			return false;
		}
		public void write(PrintStream out, String prefix) {
			if(separatorDefinition.separator.endsWith("\\"))
				out.println(prefix+"csv.separator.name="+separatorDefinition.name);
			else
				out.println(prefix+"csv.separator="+separatorDefinition.separator);
			if(quotesDefinition==null)
				out.println(prefix+"csv.quotes.name=NONE");
			else if(quotesDefinition.left.endsWith("\\")||quotesDefinition.right.endsWith("\\"))
				out.println(prefix+"csv.quotes.name="+quotesDefinition.name);
			else {
				out.println(prefix+"csv.quotes.left="+quotesDefinition.left);
				out.println(prefix+"csv.quotes.right="+quotesDefinition.right);
			}
			if(escapeDefinition==null)
				out.println(prefix+"csv.escape.name=NONE");
			else
				out.println(prefix+"csv.escape="+escapeDefinition.escape+" ");
			
		}
		
	}

	private Pattern escapeDefinition = Pattern.compile("^[ \\t]*([^ \\t]+)[ \\t]+(ESCAPE)\\t+([^\\t]).*$");
	private Pattern separatorDefinition = Pattern.compile("^[ \\t]*([^ \\t]+)[ \\t]+(SEPARATOR)(.+)$");
	private Pattern quotesDefinition = Pattern.compile("^[ \\t]*([^ \\t]+)[ \\t]+(QUOTES)\\t+([^\\t]+)\\t+([^\\t]+)(\\t.*)?$");
	private Pattern dateDefinition = Pattern.compile("^[ \\t]*([^ \\t]+)[ \\t]+(DATE|DATETIME|TIME|TIMESTAMP)\\t+([^\\t]+)\\t+([^\\t]+)\\t+([^\\t]+)(\\t.*)?$");
	private Pattern numericDefinition = Pattern.compile("^[ \\t]*([^ \\t]+)[ \\t]+(INTEGER|NUMBER)\\t+([^\\t]+)\\t+([^\\t]+)(\\t.*)?$");
	
//  .[{}()\*+?|^$	
	private List<ColumnFormat> qualifiedColumnFormats=new ArrayList<>();
	private List<EscapeDefinition> escapeDefinitions=new ArrayList<>();
	private List<SeparatorDefinition> separatorDefinitions=new ArrayList<>();
	private List<QuotesDefinition> quoteDefinitions=new ArrayList<>();
	
	private Map<String,ColumnFormat> columnFormatByName=null;
	private Map<String,EscapeDefinition> escapeDefinitionByName=null;
	private Map<String,SeparatorDefinition> separatorDefinitionByName=null;
	private Map<String,QuotesDefinition> quoteDefinitionByName=null;

	@SuppressWarnings({"unchecked", "varargs"})
	private static <T extends Named> Map<String, T> loadMap(List<T> src, T... more) {
		Map<String, T> result = new HashMap<>();
		for(T element: src)
			result.put(element.name, element);
		if(more!=null)
			for(T t:more)
				if(t!=null)
					result.put(t.name, t);
		return result;
	}
	private void ensureMaps() {
		if(columnFormatByName==null)
			columnFormatByName=loadMap(qualifiedColumnFormats,new StringFormat());
		if(escapeDefinitionByName==null)
			escapeDefinitionByName=loadMap(escapeDefinitions);
		if(separatorDefinitionByName==null)
			separatorDefinitionByName=loadMap(separatorDefinitions);
		if(quoteDefinitionByName==null)
			quoteDefinitionByName=loadMap(quoteDefinitions);
	}
	public ColumnFormat getColumnFormat(String name) {
		ensureMaps();
		return columnFormatByName.get(name);
	}
	private boolean loadFormatDefinitions(String definition) {
		Matcher matcher;
		matcher=dateDefinition.matcher(definition);
		if(matcher.matches()) {
			qualifiedColumnFormats.add(new DateFormat(matcher.group(1), BaseFormat.valueOf(matcher.group(2)), matcher.group(3),Locale.forLanguageTag(matcher.group(4)), matcher.group(5)));
			return true;
		}
		matcher=numericDefinition.matcher(definition);
		if(matcher.matches()) {
			qualifiedColumnFormats.add(new NumberFormat(matcher.group(1), BaseFormat.valueOf(matcher.group(2)), matcher.group(3), matcher.group(4)));
			return true;
		}
		matcher=separatorDefinition.matcher(definition);
		if(matcher.matches()) {
			separatorDefinitions.add(new SeparatorDefinition(matcher.group(1), matcher.group(3)));
			return true;
		}
		matcher=quotesDefinition.matcher(definition);
		if(matcher.matches()) {
			quoteDefinitions.add(new QuotesDefinition(matcher.group(1), matcher.group(3), matcher.group(4)));
			return true;
		}
		matcher=escapeDefinition.matcher(definition);
		if(matcher.matches()) {
			escapeDefinitions.add(new EscapeDefinition(matcher.group(1), matcher.group(3)));
			return true;
		}
		return false;
	}
	private void loadFormatDefinitions(BufferedReader definitions) throws IOException {
		for(String definition=definitions.readLine();definition!=null;definition=definitions.readLine()) {
			
			if (loadFormatDefinitions(definition)) {
				AbstractTool.LOG.fine("loaded definition "+definition);
			} else if(definition!= null) {
				String d = definition.trim();
				if (!(d.isEmpty()||d.startsWith("-")||d.startsWith("/")||d.startsWith("#")))
					AbstractTool.LOG.warning("unidentified definition: "+definition);
			}
		}
	}
	void loadFormatDefinitions(File configDir, Properties taskProperties) throws IOException {
		try (Reader reader=AbstractTool.fileOrResourceUTF8(configDir, taskProperties.getProperty("csv.formats"), "CSVFormatDefinitions.txt", "com/enitsys/sqltools/CSVFormatDefinitions.txt", "CSVFormatDefinitions.txt", "com/enitsys/sqltools/CSVFormatDefinitions.txt")) {
			if (reader==null)
				throw new NullPointerException("reader");
			try (BufferedReader br = new BufferedReader(reader)) {
				loadFormatDefinitions(br);
			}
		}
	}
	
	List<CSVDefinition> createAllCSVDefinitions() {
		List<CSVDefinition> result=new ArrayList<>();
		CSVDefinition csv;
		for(SeparatorDefinition separatorDefinition:separatorDefinitions) {
			csv=CSVDefinition.valueOf(separatorDefinition, null, null);
			if (csv!=null)
				result.add(csv);
			for(QuotesDefinition quotesDefinition:quoteDefinitions) {
				csv=CSVDefinition.valueOf(separatorDefinition, quotesDefinition, null);
				if (csv!=null)
					result.add(csv);
				for(EscapeDefinition escapeDefinition:escapeDefinitions) {
					csv=CSVDefinition.valueOf(separatorDefinition, quotesDefinition, escapeDefinition);
					if (csv!=null)
						result.add(csv);
				}
			}
		}
		return result;
	}
	public List<ColumnFormat> getQualifiedColumnFormats() {
		return qualifiedColumnFormats;
	}
	public static void main (String[] args) {
		CSVFormat fa=new CSVFormat();
		
		try {
			fa.loadFormatDefinitions(null, null);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return;
		}
		System.out.println(fa.qualifiedColumnFormats.size()+" FormatDefinitions loaded");
		List<CSVDefinition> csvDefList=fa.createAllCSVDefinitions();
		System.out.println("Now Testing "+csvDefList.size()+" csv Definitions ...");
		CSVDefinition[] csvDefinitions=csvDefList.toArray(new CSVDefinition[csvDefList.size()]);
		try (InputStream is = CSVFormat.class.getClassLoader().getResourceAsStream("com/enitsys/sqltools/csv/Test1.txt")) {
			if (is==null)
				throw new NullPointerException("is");
			try (LineNumberReader br = new LineNumberReader(new InputStreamReader(is,"UTF-8"))) {
				for(String line=br.readLine();line!=null;line=br.readLine()) {
					for(CSVDefinition csvDefinition: csvDefinitions) 
						if(csvDefinition!=null)
							csvDefinition.analyse(br.getLineNumber(), line);
				}
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		ArrayList<CSVDefinition> list=new ArrayList<>(csvDefinitions.length);
		for(CSVDefinition csvDefinition : csvDefinitions) 
			if(csvDefinition!=null && !csvDefinition.isRemovable(true))
				list.add(csvDefinition);
		csvDefinitions=list.toArray(new CSVDefinition[list.size()]);
		Arrays.sort(csvDefinitions);
		
		System.out.println("File might be one of the following csv Formats:");
		for(CSVDefinition csvDefinition : csvDefinitions) 
			System.out.println("  "+ csvDefinition);
	}
	
	/*
	public static void main (String[] args) {
		long START=System.currentTimeMillis();
		Pattern p=Pattern.compile("([^,]*)");
		Pattern s=Pattern.compile(",");
		
		Matcher m=p.matcher("1,2,3,,,6,,");
		while (m.lookingAt()) {
			System.out.println("P=<"+m.group(1)+">");
			if(m.end()==m.regionEnd())
				break;
			m.region(m.end(), m.regionEnd());
			m.usePattern(s);
			if(!m.lookingAt())
				throw new RuntimeException("Fehler");
			m.region(m.end(), m.regionEnd());
			m.usePattern(p);
		}
		long END=System.currentTimeMillis();
		System.out.println((END-START)+" millis");
	}
*/
	/*
	public static void main (String[] args) {
		CSVDefinition csv=new CSVDefinition(new SeparatorDefinition(null, "abc"), new EscapeDefinition(null, "$"), new BoundaryDefinition(null, "((", "))"));
		
		//Pattern p=Pattern.compile("(.*(?!abc))");
		long START=System.currentTimeMillis();
		Pattern s=Pattern.compile("abc");
		Pattern l=Pattern.compile("\\(\\(");
		Pattern r=Pattern.compile("\\)\\)");
		Pattern e=Pattern.compile("(^|[^\\$])\\$(\\$\\$)*[^\\$]");
		
//		String val="1abc((2))abc((3$$abc$)))abcabc5abcabc";
		String val="((3$$abc$)))abcabc5abcabc((x$))";
		Matcher m=l.matcher(val);
		Matcher em=e.matcher(val);
		
		int valStart=0;
		int valEnd=0;
		int escaped=Integer.MAX_VALUE;
		if(em.find()) {
			escaped=em.end()-1;
			em.region(escaped+1, em.regionEnd());
		} 
			
		boolean sepFound=false;
		int endStart=0;
		for (int i=1;;i++) {
			m.usePattern(l);
			if(m.lookingAt()) {
				valStart=m.end();
				m.region(m.end(), m.regionEnd());
				m.usePattern(r);
				for(boolean endFound=false;!endFound;) {
					endFound=m.find();
					if(!endFound)
						throw new RuntimeException("Position "+ m.regionStart() +": Closing ColumnBoundaries "+r.pattern()+" not found.");
					endStart=m.start();
					while (escaped<endStart) {
						if(em.find()) {
							escaped=em.end()-1;
							em.region(escaped+1, em.regionEnd());
						} else 
							escaped=Integer.MAX_VALUE;
					}
					if (escaped==endStart) {
						endFound=false;
						m.region(escaped+1, m.regionEnd());
					}
				}
				valEnd=endStart;
				m.region(m.end(), m.regionEnd());
				
				m.usePattern(s);
				sepFound=m.lookingAt();
				if(!sepFound && m.regionStart() < m.regionEnd() )
					throw new RuntimeException("Position "+ m.regionStart() +": EOL or ColumnSeparator "+s.pattern()+" expected.");
			} else {
				valStart=m.regionStart();
				m.usePattern(s);
				for(boolean endFound=false;!endFound;) {
					sepFound=m.find();
					if(!sepFound)
						break;
					endStart=m.start();
					while (escaped<endStart) {
						if(em.find()) {
							escaped=em.end()-1;
							em.region(escaped+1, em.regionEnd());
						} else 
							escaped=Integer.MAX_VALUE;
					}
					endFound = (escaped != endStart);
					if(!endFound)
						m.region(escaped+1, m.regionEnd());
				}
				valEnd=sepFound?m.start():m.regionEnd();
			}
			System.out.println(i+": <"+val.substring(valStart, valEnd)+">");
			if(!sepFound)
				break;
			m.region(m.end(), m.regionEnd());
		}
		long END=System.currentTimeMillis();
		System.out.println((END-START)+" millis");

	}
	*/
	/*
	public static void main (String[] args) {
		long START=System.currentTimeMillis();
		CSVDefinition csv=new CSVDefinition(new SeparatorDefinition(null, "abc"), new EscapeDefinition(null, "$"), new QuotesDefinition(null, "((", "))"));
//		String val="1abc((2))abc((3$$abc$)))abcabc5abcabc";
		String val="((3$$$$$abc$)))abcabc5abcabc((x))";

		List<String> cols=csv.getValues(val);
		for (int i=0;i<cols.size();i++)
			System.out.println(i+": <"+cols.get(i)+">");
		
		long END=System.currentTimeMillis();
		System.out.println((END-START)+" millis");

	}
	*/
}
