/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipOutputStream;

/**
 * Under Construction.
 */
public class SQL2archive extends AbstractSQLTool implements Archivable {
    private final String connection;
    private final String directory;
    public SQL2archive(String taskname, String connection, String directory) throws IOException {
        super(taskname, connection, directory, false);
        this.connection=connection;
        this.directory=directory;
    }

    @Override
    public void write(Connection dataConnection, OutputStreamFactory outputStreamFactory, boolean addSQLFile) throws SQLException, IOException {
        Pattern pattern = Pattern.compile("(SQL2archive|SQL2csv|SQL2file|SQL2xml):([^:]+)(:.+)?");
        String tasks = taskProperties.getProperty("tasks");
        String taskDef;
        for(StringTokenizer stringTokenizer=new StringTokenizer(tasks,";"); stringTokenizer.hasMoreTokens();) {
            taskDef=stringTokenizer.nextToken();
            Matcher matcher = pattern.matcher(taskDef);
            if (!matcher.matches())
                throw new RuntimeException("Taskdefinition '"+taskDef+"' is not valid.");
            String clazz = matcher.group(1);
            String taskname = matcher.group(2);
            String connection = matcher.group(3);
            Archivable task = null;
            if ("SQL2archive".equals(clazz))
                task = new SQL2archive(taskname, connection!=null?connection:this.connection, this.directory);
            else if ("SQL2csv".equals(clazz))
                task = new SQL2csv(taskname, connection!=null?connection:this.connection, this.directory);
            else if ("SQL2file".equals(clazz))
                task = new SQL2file(taskname, connection!=null?connection:this.connection, this.directory);
            else
                task = new SQL2xml(taskname, connection!=null?connection:this.connection, this.directory);
            task.write(connection==null?dataConnection:null, outputStreamFactory, addSQLFile);

        }

    }

    @Override
    public void run() throws SQLException, IOException {
        try (Connection conn = createConnection()) {
            String zipFile = replace(System.getProperty("out.zip.file", taskProperties.getProperty("out.zip.file")), replacements);
            if (zipFile != null) {
                File outfile = (destDir == null || new File(zipFile).isAbsolute()) ? new File(zipFile) : new File(destDir, zipFile);
                Files.createDirectories(outfile.toPath().getParent());
                try (OutputStreamFactory outputStreamFactory = new OutputStreamFactory.ZipEntryOutputStreamFactory(new ZipOutputStream(new FileOutputStream(outfile)))) {
                    write(conn, outputStreamFactory, false);
                }
            } else {
                String tempDir = System.getProperty("out.tempdir", taskProperties.getProperty("out.tempdir"));
                OutputStreamFactory outputStreamFactory = null;
                if (tempDir != null)
                    outputStreamFactory = new OutputStreamFactory.TempOutputStreamFactory(destDir, tempDir);
                else
                    outputStreamFactory = new OutputStreamFactory.DirOutputStreamFactory(destDir);
                write(conn, outputStreamFactory, false);
            }
        }

    }

    public static void main(String[] args) {
        main(SQL2archive.class, args);

    }

}
