/*
 * Copyright 2023, Rainer Noack, Frankfurt am Main, Germany.
 *
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. *
 */
package com.enitsys.sqltools;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class SQLScriptReader implements AutoCloseable {
    private static final char LINE_SEPARATOR = '\n';
    private static final String DEFAULT_LINE_COMMENT = "--";
    private static final Pattern DELIMITER_PATTERN = Pattern.compile("^\\s*((--)|(//))?\\s*(//)?\\s*@DELIMITER\\s+([^\\s]+)", Pattern.CASE_INSENSITIVE);
    private static final Pattern BLOCKDELIMITER_PATTERN = Pattern.compile("^\\s*((--)|(//))?\\s*(//)?\\s*@BLOCKDELIMITER\\s+([^\\s]+)", Pattern.CASE_INSENSITIVE);
    private static final Pattern BLOCKENTRY_PATTERN = Pattern.compile("^\\s*((--)|(//))?\\s*(//)?\\s*@BLOCKENTRY\\s+([^\\s]+)", Pattern.CASE_INSENSITIVE);
    private static final Pattern BLOCKCLEAR_PATTERN = Pattern.compile("^\\s*((--)|(//))?\\s*(//)?\\s*@CLEARBLOCK", Pattern.CASE_INSENSITIVE);

    private final BufferedReader reader;
    private String nextStatement;
    private boolean[] wasLast = {false};
    private final String lineComment;
    private String delimiter;
    private String blockDelimiter;
    private List<String> blockEntries = new ArrayList<>(5);
    private boolean fullLineDelimiter = false;
    private boolean blockMode = false;
    private static final String trimNullIfEmpty(String x, boolean upper) {
        if (x==null)
            return null;
        x=x.trim();
        return x.isEmpty()?null:(upper?x.toUpperCase(Locale.ROOT):x);
    }
    SQLScriptReader(BufferedReader reader, Properties scriptProperties) throws IOException {
        this.reader=reader;
        this.lineComment= scriptProperties.getProperty("script.line-comment", DEFAULT_LINE_COMMENT).trim();
        this.delimiter= trimNullIfEmpty(scriptProperties.getProperty("script.delimiter"), false);
        this.blockDelimiter=trimNullIfEmpty(scriptProperties.getProperty("script.block-delimiter"), true);
        String blockEntries=scriptProperties.getProperty("script.block-entries");
        if(blockEntries!=null) {
            String entry;
            for (StringTokenizer stringTokenizer = new StringTokenizer(blockEntries, ","); stringTokenizer.hasMoreTokens(); ) {
                entry = trimNullIfEmpty(stringTokenizer.nextToken(), true);
                if (entry != null)
                   this.blockEntries.add(entry);
            }
        }
        nextStatement=readNextStatement(wasLast);
    }

    @Override
    public void close() throws IOException {
        reader.close();
    }
    public String readStatement() throws IOException {
        if(nextStatement==null)
            return null;
        String result=nextStatement;
        nextStatement=wasLast[0]?null:readNextStatement(wasLast);
        return result;
    }

        private String readNextStatement(boolean[] wasLast) throws IOException {
            StringBuilder command = new StringBuilder();
            String line;
            while ((line = reader.readLine()) != null)
                 if (handleLine(command, line)) {
                     wasLast[0]=false;
                     return command.toString();
                 }
            wasLast[0]=true;
            return command.toString();
        }

        private boolean handleLine(StringBuilder command, String line) {
            String trimmedLine = line.trim();
            if (lineIsComment(trimmedLine)) {
                Matcher matcher = DELIMITER_PATTERN.matcher(trimmedLine);
                if (matcher.find())
                    delimiter = matcher.group(5);
                else {
                    matcher = BLOCKCLEAR_PATTERN.matcher(trimmedLine);
                    if (matcher.find())
                        blockEntries.clear();
                    else {
                        matcher = BLOCKENTRY_PATTERN.matcher(trimmedLine);
                        if (matcher.find())
                            blockEntries.add(matcher.group(5).toUpperCase(Locale.ROOT));
                        else {
                            matcher = BLOCKDELIMITER_PATTERN.matcher(trimmedLine);
                            if (matcher.find())
                                delimiter = matcher.group(5).toUpperCase(Locale.ROOT);
                        }
                    }
                }
                if (command.length()>0)
                    command.append(LINE_SEPARATOR);
                command.append(line);
            } else {
                if (lineStartsBlock(trimmedLine)) {
                    blockMode = true;
                    command.delete(0,command.length());
                }
                if (commandReadyToExecute(trimmedLine)) {
                    if (command.length()>0)
                        command.append(LINE_SEPARATOR);
                    if (blockMode)
                        blockMode = false;
                    else
                        command.append(line, 0, line.lastIndexOf(delimiter));
                    return true;
                } else if (trimmedLine.length() > 0 || command.length() > 0) {
                    if (command.length()>0)
                        command.append(LINE_SEPARATOR);
                    command.append(line);
                }
            }
            return false;
        }

        private boolean lineIsComment(String trimmedLine) {
            return trimmedLine.startsWith(lineComment);
        }
        private boolean lineStartsBlock(String trimmedLine) {
            if (blockMode)
                return false;
            if(trimmedLine.isEmpty())
                return false;
            if (delimiter==null && blockEntries.isEmpty())
                return true;
            String upperLine=trimmedLine.toUpperCase(Locale.ROOT);
            for(String blockEntry : blockEntries)
                if (upperLine.startsWith(blockEntry))
                    return true;
            return false;
        }

        private boolean commandReadyToExecute(String trimmedLine) {
            if(trimmedLine.isEmpty())
                return false;
            if(blockMode) {
                if(blockDelimiter==null)
                    return false;
                return trimmedLine.toUpperCase(Locale.ROOT).startsWith(blockDelimiter);
            }
            if (delimiter==null)
                return false;
            return (!fullLineDelimiter && trimmedLine.contains(delimiter)) || (fullLineDelimiter && trimmedLine.equals(delimiter));
        }

}
